/**
 * control HD44780 compatible LCD-display via serial commands
 * 
 * This program controls a HD44780 compatible LCD-display via an UART
 * connection. It is used to connect the display to the 1-wire bus via
 * another controller which acts as a 1-wire slave.
 * FUSES for the ATTiny2313: BrownOutDetection at 4.3V, CKOUT enable,
 * high speed crystal oscillator startup 14CK + 65ms
 * 
 * ----------------------------------------------------------------------------------
 * 
 * Copyright (C) 2009 runlevel3 GmbH
 * 
 * ----------------------------------------------------------------------------------
 * 
 * LICENSE:
 * 
 * This file is part of AVR-projects.
 * 
 * AVR-projects is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * AVR-projects is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with AVR-projects; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 * ----------------------------------------------------------------------------------
 * 
 * AVR Assembler 2
 * 
 * ----------------------------------------------------------------------------------
 * 
 * @copyright 2009 runlevel3 GmbH
 * @license   http://www.gnu.org/licenses/gpl.txt GNU General Public License v2
 * @link      http://www.runlevel3.de runlevel3
 **/



.include "tn2313def.inc"





#define	F_CPU		20000000




;register definitions

#ifndef	TMP_REGS

#define	TMP_REGS

.def	tmp0		= R24

.def	tmp1		= R25

#endif

#ifndef	INT_REGS

#define	INT_REGS

.def	int_sr		= R15
.def	int_tmp		= R21
#endif


.def	timer_10ms	= R16
.def	lcd_pwm		= R17




.equ	DATA_BUFFER_SIZE	= 33			;1 byte address and 32 byte data
.equ	CMD_TIMEOUT			= 10			;10*10ms timeout -> 100ms


;device commands w1_devcmd

.equ	DC_LCD_CLEAR		= 0x1f

.equ	DC_LCD_READ			= 0xab

.equ	DC_LCD_WRITE		= 0x56

.equ	DC_LCD_RAW_READ		= 0xf6

.equ	DC_LCD_RAW_WRITE	= 0x16

.equ	DC_LCD_PWM			= 0x6a



.equ	STX					= 0x02

.equ	ETX					= 0x03





;timer0 phase correct PWM, prescaler: 64

.equ	T0CA				= (1<<WGM00)|(1<<COM0B1)
.equ	T0CB				= (1<<CS01)|(1<<CS00)

;timer1 10ms, prescaler = 8, clear on compare match OCR1A
.equ	T1_10MS				= 25000
.equ	T1CB				= (1<<CS01)|(1<<WGM12)

.cseg										;Code-Segment

.org	0

 		rjmp 	reset



.org	OC1Aaddr
		rjmp	ISR_OC1A					;timer1 compare match A interrupt

.org	INT_VECTORS_SIZE					;reserve space for other int vectors



.include	"uart-int-fd-buff.asm"


.org		uart_int_fd_buff_cseg_end

.include	"lcd-alpha.asm"


.org		lcd_alpha_cseg_end

#define	CRC_COMPUTE							;do not include tables for crc, compute
.include	"crc.asm"


.org		crc_cseg_end




.dseg

w1_lcd_cmd:	.byte	1

w1_lcd_num:	.byte	1

w1_lcd_addr:.byte	1

w1_data_buf:.byte	DATA_BUFFER_SIZE



.cseg										;Code-Segment


ISR_OC1A:
		in		int_sr,SREG
		inc		timer_10ms					;cmd timeout
		in		int_tmp,OCR0B				;pwm change?
		cp		int_tmp,lcd_pwm
		breq	pwm_nochange				;no, exit
		brcc	pwm_down
		inc		int_tmp						;pwm up
		rjmp	pwm_out
pwm_down:
		dec		int_tmp						;pwm down
pwm_out:
		out		OCR0B,int_tmp				;set pwm
pwm_nochange:
		out		SREG,int_sr
		reti



reset:

		ldi		tmp0,1<<CLKPCE				;Clock prescaler change enable

		out		CLKPR,tmp0

		clr		tmp0						;no prescaler

		out		CLKPR,tmp0

		ldi		tmp0,RAMEND					;init stackpointer
		out		SPL,tmp0

		ldi		tmp0,T1CB					;load prescaler and control timer1
		out		TCCR1B,tmp0
		ldi		tmp0,HIGH(T1_10MS)			;load 10ms value
		out		OCR1AH,tmp0
		ldi		tmp0,LOW(T1_10MS)			;load 10ms value
		out		OCR1AL,tmp0
		
		ldi		tmp0,T0CA					;load prescaler and control timer0
		out		TCCR0A,tmp0
		ldi		tmp0,T0CB
		out		TCCR0B,tmp0
		ser		tmp0						;pwm max
		out		OCR0B,tmp0
		sbi		DDRD,PD5					;OC0B Pin PWM Output

		in		tmp0,TIMSK					;enable timer1 OC1A interrupt
		ori		tmp0,(1<<OCIE1A)
		out		TIMSK,tmp0

		rcall	int_uart_init				;init full duplex uart
		rcall	lcd_init					;init display
		sbi		LCD_PORT,LCD_UNUSED			;unused pin as activity led -> led off
		sbi		LCD_DDR,LCD_UNUSED
		ldi		ZH,HIGH(2*init_text)
		ldi		ZL,LOW(2*init_text)
		rcall	lcd_put_PGM_str

		sei

		ldi		ZH,HIGH(2*init_text)
		ldi		ZL,LOW(2*init_text)
		rcall	int_uart_put_PGM_str
		rcall	int_uart_put_enter

main:
		sbi		LCD_PORT,LCD_UNUSED			;led off
state_idle:
		rcall	cb_rx_used					;bytes in buffer?
		breq	state_idle					;no -> loop
		cbi		LCD_PORT,LCD_UNUSED			;led on
		rcall	int_uart_getchar			;get char from buffer
		cpi		tmp0,STX					;start sequence?
		brne	main						;no -> loop
state_cmd:
		rcall	cb_rx_used					;bytes in buffer?
		breq	state_cmd					;no -> wait
		rcall	int_uart_getchar			;get char from buffer
		cpi		tmp0,STX
		breq	state_cmd
		clr		tmp1						;init crc
		sts		crc16_data_h,tmp1
		sts		crc16_data_l,tmp1
		rcall	crc16						;cmd -> crc
		sts		w1_lcd_cmd,tmp0

		cpi		tmp0,DC_LCD_CLEAR
		breq	state_num
		cpi		tmp0,DC_LCD_READ
		breq	state_num
		cpi		tmp0,DC_LCD_WRITE

		breq	state_num
		cpi		tmp0,DC_LCD_RAW_READ
		breq	state_num
		cpi		tmp0,DC_LCD_RAW_WRITE
		breq	state_num
		cpi		tmp0,DC_LCD_PWM
		breq	state_num
		rjmp	main						;unknown command -> idle

state_num:
		clr		timer_10ms					;observe timeout
state_num_loop:
		cpi		timer_10ms,CMD_TIMEOUT+1
		brsh	main						;timeout
		rcall	cb_rx_used					;bytes in buffer?
		breq	state_num_loop				;no -> wait
		rcall	int_uart_getchar			;get char from buffer
		rcall	crc16						;num -> crc
		cpi		tmp0,DATA_BUFFER_SIZE+1		;max DATA_BUFFER_SIZE (40) byte data?
		brsh	main						;too big data idle

state_data:
		ldi		XH,HIGH(w1_data_buf)		;load start address data buffer
		ldi		XL,LOW(w1_data_buf)
		mov		tmp1,tmp0					;load to loopcounter
		sts		w1_lcd_num,tmp1				;and save for crc loopcount
		tst		tmp1						;packet with 0 byte data?
		breq	state_crc
state_data_loop:
		cpi		timer_10ms,CMD_TIMEOUT+1
		brsh	main						;timeout
		rcall	cb_rx_used					;bytes in buffer?
		breq	state_data_loop				;no -> wait
		rcall	int_uart_getchar			;get char from buffer
		rcall	crc16						;data -> crc
		st		X+,tmp0						;store in data buffer
		dec		tmp1						;more data?
		brne	state_data_loop
state_crc:
		cpi		timer_10ms,CMD_TIMEOUT+1
		brsh	main						;timeout
		rcall	cb_rx_used					;bytes in buffer?
		cpi		tmp0,2						;waiting for two bytes crc
		brlo	state_crc					;no -> wait
		rcall	int_uart_getchar			;get crc_l from buffer
		lds		tmp1,crc16_data_l			;check crc
		cpse	tmp0,tmp1
		rjmp	main						;wrong crc -> idle

		rcall	int_uart_getchar			;get crc_h from buffer
		lds		tmp1,crc16_data_h			;check crc
		cpse	tmp0,tmp1
		rjmp	main						;wrong crc -> idle

state_etx:
		cpi		timer_10ms,CMD_TIMEOUT+1
		brsh	state_exit					;timeout
		rcall	cb_rx_used					;bytes in buffer?
		breq	state_etx					;no -> wait
		rcall	int_uart_getchar			;get char from buffer
		cpi		tmp0,ETX					;start sequence?
		brne	state_exit					;no -> loop
		rcall	lcd_exec
state_exit:
		rjmp	main



;execute crc-checked command in w1_lcd_cmd with length w1_lcd_num in w1_data_buf
lcd_exec:
		ldi		XH,HIGH(w1_data_buf)		;load start address data buffer
		ldi		XL,LOW(w1_data_buf)
		lds		tmp0,w1_lcd_cmd

		cpi		tmp0,DC_LCD_CLEAR
		breq	lcd_exec_init
		cpi		tmp0,DC_LCD_READ
		breq	lcd_exec_read
		cpi		tmp0,DC_LCD_WRITE

		breq	lcd_exec_write
		cpi		tmp0,DC_LCD_RAW_READ
		breq	lcd_exec_raw_read_tmp
		cpi		tmp0,DC_LCD_RAW_WRITE
		breq	lcd_exec_raw_write_tmp
		cpi		tmp0,DC_LCD_PWM
		breq	lcd_exec_pwm
		ret
lcd_exec_raw_read_tmp:
		rjmp	lcd_exec_raw_read
lcd_exec_raw_write_tmp:
		rjmp	lcd_exec_raw_write

lcd_exec_init:
		rcall	lcd_clear
		clr		tmp0						;return 0 data bytes
		sts		w1_lcd_num,tmp0
		rcall	lcd_send_result
		ret

lcd_exec_pwm:
		ld		lcd_pwm,X					;address to read from
		clr		tmp0						;return 0 data bytes
		sts		w1_lcd_num,tmp0
		rcall	lcd_send_result
		ret

lcd_exec_read:
		ld		tmp0,X						;address to read from
		andi	tmp0,0xc0					;cg or dd ram address?
		brne	lcd_exec_read_start
		ser		tmp0						;return -1 -> error
		sts		w1_lcd_cmd,tmp0
		clr		tmp0						;return 0 data bytes
		sts		w1_lcd_num,tmp0
		rjmp	lcd_exec_read_exit
lcd_exec_read_start:
		ld		tmp0,X+						;address to read from
		clr		tmp1						;RS=0 -> send command
		rcall	lcd_byte_out
		ld		tmp1,X						;load loopcounter
		ldi		XH,HIGH(w1_data_buf)		;load start address data buffer
		ldi		XL,LOW(w1_data_buf)
		cpi		tmp1,DATA_BUFFER_SIZE		;max DATA_BUFFER_SIZE-1 (32) byte data?
		brlo	lcd_exec_read_size			;data size ok

		ldi		tmp1,DATA_BUFFER_SIZE-1		;load max DATA_BUFFER_SIZE
lcd_exec_read_size:
		sts		w1_lcd_num,tmp1				;store data size for result sending
		tst		tmp1						;0 byte to read?
		breq	lcd_exec_read_exit
lcd_exec_read_loop:
		push	tmp1						;save loopcounter
		ldi		tmp1,1<<LCD_RS				;RS=1 -> read data
		rcall	lcd_byte_in
		st		X+,tmp0
		pop		tmp1						;restore loopcounter
		dec		tmp1
		brne	lcd_exec_read_loop
lcd_exec_read_exit:
		rcall	lcd_send_result
		ret

lcd_exec_write:
		ld		tmp0,X						;address for write to
		andi	tmp0,0xc0					;cg or dd ram address?
		brne	lcd_exec_write_start
		ser		tmp0						;return -1 -> error
		sts		w1_lcd_cmd,tmp0
		rjmp	lcd_exec_write_exit
lcd_exec_write_start:
		ld		tmp0,X+						;address for write to
		clr		tmp1						;RS=0 -> send command
		rcall	lcd_byte_out
		lds		tmp1,w1_lcd_num				;load command length
		dec		tmp1						;sub 1 for address
		breq	lcd_exec_write_exit			;no data to send
lcd_exec_write_loop:
		push	tmp1						;save loopcounter
		ldi		tmp1,1<<LCD_RS				;RS=1 -> send data
		ld		tmp0,X+
		rcall	lcd_byte_out
		pop		tmp1						;restore loopcounter
		dec		tmp1
		brne	lcd_exec_write_loop
lcd_exec_write_exit:
		clr		tmp0						;return 0 data bytes
		sts		w1_lcd_num,tmp0
		rcall	lcd_send_result
		ret

lcd_exec_raw_read:
		clr		tmp1						;RS=0 -> read command
		rcall	lcd_byte_in
		st		X,tmp0						;store at buffer
		ldi		tmp0,1						;return 1 data bytes
		sts		w1_lcd_num,tmp0
		rcall	lcd_send_result
		ret

lcd_exec_raw_write:
		lds		tmp1,w1_lcd_num				;load data length
lcd_exec_raw_write_loop:
		push	tmp1						;save loopcounter
		clr		tmp1						;RS=0 -> send command
		ld		tmp0,X+						;load data
		rcall	lcd_byte_out
		pop		tmp1						;restore loopcounter
		dec		tmp1
		brne	lcd_exec_raw_write_loop
		clr		tmp0						;return 0 data bytes
		sts		w1_lcd_num,tmp0
		rcall	lcd_send_result
		ret



lcd_send_result:
		rcall	cb_tx_free
		cpi		tmp0,3
		brmi	lcd_send_result
		ldi		tmp0,STX					;send start byte
		rcall	int_uart_putchar
		clr		tmp0						;init crc
		sts		crc16_data_l,tmp0
		sts		crc16_data_h,tmp0
		lds		tmp0,w1_lcd_cmd				;send last command or error code
		rcall	crc16
		rcall	int_uart_putchar
		lds		tmp0,w1_lcd_num				;send num bytes back
		mov		tmp1,tmp0					;load loopcounter
		rcall	crc16
		rcall	int_uart_putchar
		tst		tmp1						;packet with 0 byte data?
		breq	lcd_send_crc
		cpi		tmp1,DATA_BUFFER_SIZE		;max DATA_BUFFER_SIZE (32) byte data?
		brsh	lcd_send_crc				;too big data just send crc

		ldi		XH,HIGH(w1_data_buf)		;load start address data buffer
		ldi		XL,LOW(w1_data_buf)
lcd_send_data_loop:
		rcall	cb_tx_free					;wait fro free buffer
		cpi		tmp0,1
		brmi	lcd_send_data_loop
		ld		tmp0,X+						;load data from buffer
		rcall	crc16						;data -> crc
		rcall	int_uart_putchar			;send out
		dec		tmp1						;more data?
		brne	lcd_send_data_loop
lcd_send_crc:
		rcall	cb_tx_free					;wait fro free buffer
		cpi		tmp0,3
		brmi	lcd_send_crc
		lds		tmp0,crc16_data_l			;send crc
		rcall	int_uart_putchar
		lds		tmp0,crc16_data_h
		rcall	int_uart_putchar
		ldi		tmp0,ETX					;end byte
		rcall	int_uart_putchar
		ret



init_text:
		.db		"runlevel3 w1-LCD",0,0
