/**
 * functions for LCD-displays with HD44780 and compatible controllers
 * 
 * Connection is in 4-bit mode and uses 7 bits of a PORT of the controller. The
 * BUSY-FLAG is used to synchronize bus transfers.
 * 
 * ----------------------------------------------------------------------------------
 * 
 * Copyright (C) 2009 runlevel3 GmbH
 * 
 * ----------------------------------------------------------------------------------
 * 
 * LICENSE:
 * 
 * This file is part of AVR-projects.
 * 
 * AVR-projects is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * AVR-projects is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with AVR-projects; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 * ----------------------------------------------------------------------------------
 * 
 * AVR Assembler 2
 * 
 * ----------------------------------------------------------------------------------
 * 
 * @copyright 2009 runlevel3 GmbH
 * @license   http://www.gnu.org/licenses/gpl.txt GNU General Public License v2
 * @link      http://www.runlevel3.de runlevel3
 **/



.include	"delay.asm"



;register definitions
#ifndef	TMP_REGS

#define	TMP_REGS

.def	tmp0		= R24

.def	tmp1		= R25

#endif


#define LCD_DATA		(1<<PB0)|(1<<PB1)|(1<<PB2)|(1<<PB3)
#define LCD_READ		PB4
#define LCD_E			PB5
#define LCD_RS			PB6
#define LCD_UNUSED		PB7
#define LCD_PORT		PORTB
#define LCD_DDR			DDRB
#define LCD_PIN			PINB

#define LCD_BUSY		7

#define	LCD_CMD_DELAY	50				;50us command delay
#define	LCD_CMD_LONG	5000			;5ms command delay long



;data in tmp0 preserved, RS in tmp1 preserved
lcd_nibble_out:
		push	tmp1					;save RS
		push	tmp0					;save data
		in		tmp0,LCD_PORT			;read port status
		andi	tmp0,1<<LCD_UNUSED		;filter LCD bits
		sbrc	tmp1,LCD_RS
		sbr		tmp0,1<<LCD_RS
		pop		tmp1					;restore data
		push	tmp1					;save data
		andi	tmp1,LCD_DATA			;filter data bits
		or		tmp0,tmp1
		out		LCD_PORT,tmp0
		sbi		LCD_PORT,LCD_E
		pop		tmp0
		pop		tmp1
		cbi		LCD_PORT,LCD_E
		ret

;read nibble and or with low nibble tmp0, registers preserved
lcd_nibble_in:
		push	tmp1					;save RS
		push	tmp0					;save data
		in		tmp0,LCD_PORT			;read port status
		andi	tmp0,1<<LCD_UNUSED		;filter LCD bits
		sbrc	tmp1,LCD_RS
		sbr		tmp0,1<<LCD_RS
		ori		tmp0,1<<LCD_READ		;set read bit
		out		LCD_PORT,tmp0
		sbi		LCD_PORT,LCD_E
		nop
		nop
		nop
		nop
		in		tmp1,LCD_PIN
		andi	tmp1,LCD_DATA
		pop		tmp0
		andi	tmp0,~(LCD_DATA)
		or		tmp0,tmp1
		pop		tmp1
		cbi		LCD_PORT,LCD_E
		ret

;data in tmp0, RS in tmp1, registers preserved
lcd_byte_out:
		push	tmp0					;save data
		in		tmp0,LCD_DDR			;set port I/O
		ori		tmp0,~(1<<LCD_UNUSED)
		out		LCD_DDR,tmp0
		pop		tmp0					;restore data
		swap	tmp0					;high nibble first
		rcall	lcd_nibble_out
		swap	tmp0					;low nibble
		rcall	lcd_nibble_out
		rcall	lcd_wait_busy			;wait for lcd ready
		ret

;data in tmp0, registers preserved
lcd_byte_in_nobusy:
		in		tmp0,LCD_DDR			;set port I/O
		ori		tmp0,~(1<<LCD_UNUSED)
		andi	tmp0,~(LCD_DATA)
		out		LCD_DDR,tmp0
		rcall	lcd_nibble_in			;read high nibble first
		swap	tmp0
		rcall	lcd_nibble_in			;low nibble
		ret

;data in tmp0, registers preserved
lcd_byte_in:
		push	tmp1					;save RS
		rcall	lcd_byte_in_nobusy
		clr		tmp1					;use busy flag
		rcall	lcd_wait_busy
		pop		tmp1					;restore RS
		ret

;read and wait for busy flag or do delay loops, registers preserved
lcd_wait_busy:
		push	tmp0					;save data
		push	tmp1					;save data
		sbrc	tmp1,LCD_BUSY			;dont use busy flag?
		rjmp	lcd_wait_delay
lcd_wait_busy_loop:
		clr		tmp1
		rcall	lcd_byte_in_nobusy
		sbrc	tmp0,LCD_BUSY
		rjmp	lcd_wait_busy_loop
		rjmp	lcd_wait_exit
lcd_wait_delay:
		sbrc	tmp1,LCD_RS
		rjmp	lcd_wait_delay_short	;RS=1 -> data -> short delay
		cpi		tmp0,1					;RS=0, data=1/2 -> long delay
		breq	lcd_wait_delay_long
		cpi		tmp0,2
		breq	lcd_wait_delay_long
lcd_wait_delay_short:
		clr		tmp1					;setup delay 50us
		ldi		tmp0,LCD_CMD_DELAY
		rcall	delay_us
		rjmp	lcd_wait_exit
lcd_wait_delay_long:
		ldi		tmp1,HIGH(LCD_CMD_LONG)	;setup delay 5ms
		ldi		tmp0,LOW(LCD_CMD_LONG)
		rcall	delay_us
lcd_wait_exit:
		pop		tmp1
		pop		tmp0
		ret

;put string in PGM at Z+ to display
lcd_put_PGM_str:
		push	tmp0
		push	tmp1
		ldi		tmp1,1<<LCD_RS			;send data
lcd_put_PGM_str_loop:
		lpm		tmp0,Z+					;load char
		tst		tmp0					;string terminated?
		breq	lcd_put_PGM_str_exit
		rcall	lcd_byte_out
		rjmp	lcd_put_PGM_str_loop
lcd_put_PGM_str_exit:
		pop		tmp1
		pop		tmp0
		ret

;put string in RAM at Z+ to display
lcd_put_str:
		push	tmp0
		push	tmp1
		ldi		tmp1,1<<LCD_RS			;send data
		ld		tmp0,Z+					;load char
		tst		tmp0					;string terminated?
		breq	lcd_put_str_exit
		rcall	lcd_byte_out
		rjmp	lcd_put_str
lcd_put_str_exit:
		pop		tmp1
		pop		tmp0
		ret

;read data from dd/cg ram, store at Z+, numbytes at tmp0
lcd_get_str:
		push	tmp0
		push	tmp1
		ldi		tmp1,1<<LCD_RS			;receive data
lcd_get_str_loop:
		push	tmp0					;save numbytes
		tst		tmp0
		breq	lcd_get_str_exit
		rcall	lcd_byte_in
		st		Z+,tmp0					;store char
		pop		tmp0
		dec		tmp0					;dec loop counter
		rjmp	lcd_get_str_loop
lcd_get_str_exit:
		pop		tmp0					;clean stack
		pop		tmp1
		pop		tmp0
		ret

;init 4-bit mode, clear ram, cursor home
lcd_init:
		push	ZL
		push	ZH
		push	tmp0
		push	tmp1
		in		tmp0,LCD_DDR			;set port I/O
		ori		tmp0,~(1<<LCD_UNUSED)
		out		LCD_DDR,tmp0
		ldi		tmp1,HIGH(LCD_CMD_LONG)	;setup delay 5ms
		ldi		tmp0,LOW(LCD_CMD_LONG)
		rcall	delay_us				;wait for stable interface
		ldi		ZH,high(2*init_sequence)
		ldi		ZL,low(2*init_sequence)
		ldi		tmp1,1<<LCD_BUSY		;RS=0, dont use busy flag
lcd_init_loop:
		lpm		tmp0,Z+					;load char
		tst		tmp0					;string terminated?
		breq	lcd_init_exit
		rcall	lcd_byte_out
		rjmp	lcd_init_loop
lcd_init_exit:
		pop		tmp1
		pop		tmp0
		pop		ZH
		pop		ZL
		ret

init_sequence:
		.db		0x33, 0x33, 0x32, 0x28, 6, 0xe, 0x1c, 1, 2, 0


;place cursor at row tmp0, column tmp1
lcd_goto:
		push	tmp0
		push	tmp1
		lsl		tmp0					;shift row 6 bits left
		lsl		tmp0
		lsl		tmp0
		lsl		tmp0
		lsl		tmp0
		lsl		tmp0
		or		tmp0,tmp1				;add column
		ori		tmp0,0x80				;command set ddram address
		clr		tmp1					;RS=0 -> send command
		rcall	lcd_byte_out
		pop		tmp1
		pop		tmp0
		ret

;clear display and cursor home
lcd_clear:
		push	tmp0
		push	tmp1
		ldi		tmp0,1					;command clear display
		clr		tmp1					;RS=0 -> send command
		rcall	lcd_byte_out
		ldi		tmp0,2					;command return home
		rcall	lcd_byte_out
		pop		tmp1
		pop		tmp0
		ret



lcd_alpha_cseg_end:						;cseg end address
