/**
 * interrupt driven uart, full-duplex, circular buffers
 * 
 * This file implements a full duplex uart with AVR USART hardware. Circular
 * buffers are used for receive and transmit data.
 * 
 * ----------------------------------------------------------------------------------
 * 
 * Copyright (C) 2009 runlevel3 GmbH
 * 
 * ----------------------------------------------------------------------------------
 * 
 * LICENSE:
 * 
 * This file is part of AVR-projects.
 * 
 * AVR-projects is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * AVR-projects is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with AVR-projects; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 * ----------------------------------------------------------------------------------
 * 
 * AVR Assembler 2
 * 
 * ----------------------------------------------------------------------------------
 * 
 * @copyright 2009 runlevel3 GmbH
 * @license   http://www.gnu.org/licenses/gpl.txt GNU General Public License v2
 * @link      http://www.runlevel3.de runlevel3
 **/



;register definitions
#ifndef	TMP_REGS

#define	TMP_REGS

.def	tmp0		= R24

.def	tmp1		= R25

#endif

#ifndef	INT_REGS

#define	INT_REGS

.def	int_sr		= R15
.def	int_tmp		= R21
#endif


.equ	doublebaud	= 0

.equ	baud300		= (F_CPU /    4800) * (doublebaud + 1) - 1

.equ	baud600		= (F_CPU /    9600) * (doublebaud + 1) - 1

.equ	baud1200	= (F_CPU /   19200) * (doublebaud + 1) - 1

.equ	baud2400	= (F_CPU /   38400) * (doublebaud + 1) - 1

.equ	baud4800	= (F_CPU /   76800) * (doublebaud + 1) - 1

.equ	baud9600	= (F_CPU /  153600) * (doublebaud + 1) - 1

.equ	baud14400	= (F_CPU /  230400) * (doublebaud + 1) - 1

.equ	baud19200	= (F_CPU /  307200) * (doublebaud + 1) - 1

.equ	baud28800	= (F_CPU /  460800) * (doublebaud + 1) - 1

.equ	baud38400	= (F_CPU /  614400) * (doublebaud + 1) - 1

.equ	baud57600	= (F_CPU /  921600) * (doublebaud + 1) - 1

.equ	baud76800	= (F_CPU / 1228800) * (doublebaud + 1) - 1

.equ	baud115200	= (F_CPU / 1843200) * (doublebaud + 1) - 1

.equ	baud230400	= (F_CPU / 3686400) * (doublebaud + 1) - 1

;.equ	baud		= baud115200		;9.85 but no rounding

.equ	baud		= 10				;-> define precalculated constant instead



.equ	rtx_buff_size	= 16

.dseg
rx_start:
rx_buffer:	.byte	rtx_buff_size
rx_end:
tx_start:
tx_buffer:	.byte	rtx_buff_size
tx_end:
rx_head:	.byte	1
rx_tail:	.byte	1
tx_head:	.byte	1
tx_tail:	.byte	1

.cseg									;Code-Segment

;initialize buffer and pointer
cb_init:
		push	tmp0					;save registers
		ldi		tmp0,rx_start			;init pointer
		sts		rx_tail,tmp0
		sts		rx_head,tmp0
		ldi		tmp0,tx_start
		sts		tx_tail,tmp0
		sts		tx_head,tmp0
		pop		tmp0
		ret

;return number of used bytes in buffer in tmp0
cb_rx_used:
		push	tmp1					;save registers
		lds		tmp0,rx_head
		lds		tmp1,rx_tail
		rjmp	cb_rtx_used
cb_tx_used:
		push	tmp1					;save registers
		lds		tmp0,tx_head
		lds		tmp1,tx_tail
cb_rtx_used:
		sub		tmp0,tmp1
		brpl	cb_rtx_used_exit
		subi	tmp0,-rtx_buff_size		;negativ -> addiere buffer_size
cb_rtx_used_exit:
		pop		tmp1
		ret

;return number of free bytes in buffer in tmp0
cb_rx_free:
		rcall	cb_rx_used
		rjmp	cb_rtx_free
cb_tx_free:
		rcall	cb_tx_used
cb_rtx_free:
		neg		tmp0					;free = buffer_size - used
		subi	tmp0,-(rtx_buff_size-1)	;add buffer_size - 1
		ret

;put byte in tmp0 in buffer
cb_rx_put:
		push	tmp0					;save data byte
		rcall	cb_rx_free				;is buffer space available?
		brne	cb_rx_put_space
		pop		tmp0
		clr		tmp0					;signal error
		ret
cb_rx_put_space:
		pop		tmp0					;restore data
		push	XL
		push	XH
		clr		XH
		lds		XL,rx_head
		st		X+,tmp0					;store in buffer
		cpi		XL,rx_end				;pointer at end address?
		brne	cb_rx_put_nowrap
		ldi		XL,rx_start				;load start address
cb_rx_put_nowrap:
		sts		rx_head,XL
		pop		XH
		pop		XL
		ser		tmp0					;signal ok
		ret

;put byte in tmp0 in buffer
cb_tx_put:
		push	tmp0					;save data byte
		rcall	cb_tx_free				;is buffer space available?
		brne	cb_tx_put_space
		pop		tmp0
		clr		tmp0					;signal error
		ret
cb_tx_put_space:
		pop		tmp0					;restore data
		push	XL
		push	XH
		push	tmp1
		in		tmp1,UCSRB				;stop interupts to prevent pointer corruption
		cbi		UCSRB,UDRIE
		clr		XH
		lds		XL,tx_head
		st		X+,tmp0					;store in buffer
		cpi		XL,tx_end				;pointer at end address?
		brne	cb_tx_put_nowrap
		ldi		XL,tx_start				;load start address
cb_tx_put_nowrap:
		sts		tx_head,XL
		out		UCSRB,tmp1				;stop interupts to prevent pointer corruption
		pop		tmp1
		pop		XH
		pop		XL
		ser		tmp0					;signal ok
		ret

;get byte out of buffer, return in tmp0
cb_rx_get:
int_uart_getchar:						;alias for easier programming
		rcall	cb_rx_used				;is buffer space in use?
		brne	cb_rx_get_start
		ret
cb_rx_get_start:
		push	XL
		push	XH
		push	tmp1
		in		tmp1,UCSRB				;stop interupts to prevent pointer corruption
		cbi		UCSRB,RXCIE
		clr		XH
		lds		XL,rx_tail
		ld		tmp0,X+					;load from buffer
		cpi		XL,rx_end				;pointer at end address?
		brne	cb_rx_get_nowrap
		ldi		XL,rx_start				;load start address
cb_rx_get_nowrap:
		sts		rx_tail,XL
		out		UCSRB,tmp1				;stop interupts to prevent pointer corruption
		pop		tmp1
		pop		XH
		pop		XL
		ret

;get byte out of buffer
cb_tx_get:
		rcall	cb_tx_used				;is buffer space in use?
		brne	cb_tx_get_start
		ret
cb_tx_get_start:
		push	XL
		push	XH
		clr		XH
		lds		XL,tx_tail
		ld		tmp0,X+					;load from buffer
		cpi		XL,tx_end				;pointer at end address?
		brne	cb_tx_get_nowrap
		ldi		XL,tx_start				;load start address
cb_tx_get_nowrap:
		sts		tx_tail,XL
		pop		XH
		pop		XL
		ret


;receive interrupt
ISR_URXC:
		in		int_sr,SREG
		push	tmp0
		push	tmp1
		rcall	cb_rx_free				;is buffer space available?
		breq	int_urxc_exit
		in		tmp0,UDR
		rcall	cb_rx_put				;put byte in receive buffer
int_urxc_exit:
		pop		tmp1
		pop		tmp0
		out		SREG,int_sr
		reti

;uart data register empty interrupt
ISR_UDRE:
		in		int_sr,SREG
		push	tmp0
		push	tmp1
		rcall	cb_tx_used				;bytes in sendbuffer?
		brne	int_udre_start
		cbi		UCSRB,UDRIE				;no, stop sender
		rjmp	int_udre_exit
int_udre_start:
		rcall	cb_tx_get				;get byte from receive buffer
		out		UDR,tmp0				;send out
int_udre_exit:
		pop		tmp1
		pop		tmp0
		out		SREG,int_sr
		reti

;init pins timer and interrupts
int_uart_init:
		push	tmp0					;save registers
		rcall	cb_init					;init pointers for circular buffers
		ldi		tmp0,high(baud)

		out		UBRRH,tmp0

		ldi		tmp0,low(baud)

		out		UBRRL,tmp0

		ldi		tmp0,(doublebaud<<U2X)	;set double Baudrate bit

		out		UCSRA,tmp0

		ldi		tmp0,(1<<RXEN)|(1<<TXEN)|(1<<RXCIE)	;enable receiver, transmitter and receiver interrupt

		out		UCSRB,tmp0

		ldi		tmp0,(1<<USBS)|(3<<UCSZ0)	;set frame format: 8data, 2stop bit

		out		UCSRC,tmp0
		pop		tmp0
		ret


;put char in transmit ring buffer, start sender
int_uart_putchar:
		rcall	cb_tx_put				;put byte to sendbuffer
		brne	int_uart_putchar_start
		ret
int_uart_putchar_start:
		sbi		UCSRB,UDRIE				;start sender
		ret


;output byte as hex
int_uart_putchar_hex:
		push	tmp1					;save registers
		push	tmp0
		push	tmp0					;save byte to send
		ldi		tmp1,'0'
		mov		tmp0,tmp1
		rcall	int_uart_putchar
		ldi		tmp1,'x'
		mov		tmp0,tmp1
		rcall	int_uart_putchar
		pop		tmp1					;restore byte
		push	tmp1					;save for second nibble
		swap	tmp1
		rcall	int_uart_putchar_hex_out
		pop		tmp1					;restore byte
		rcall	int_uart_putchar_hex_out
		pop		tmp0
		pop		tmp1
		ret
int_uart_putchar_hex_out:
		andi	tmp1,0x0f
		cpi		tmp1,10
		brlo	int_uart_putchar_hex_num
		subi	tmp1,-('a'-10)
		rjmp	int_uart_putchar_hex_alpha
int_uart_putchar_hex_num:
		subi	tmp1,-'0'
int_uart_putchar_hex_alpha:
		mov		tmp0,tmp1
		rcall	int_uart_putchar
		ret

;output enter
int_uart_put_enter:
		push	tmp0					;save registers
int_uart_put_enter_loop:
		rcall	cb_tx_free
		cpi		tmp0,2
		brmi	int_uart_put_enter_loop
		ldi		tmp0,13
		rcall	int_uart_putchar
		ldi		tmp0,10
		rcall	int_uart_putchar
		pop		tmp0
		ret

;put string in pgm memory (Z)
int_uart_put_PGM_str:
		push	tmp0					;save registers
int_uart_put_PGM_str_loop:
		rcall	cb_tx_free
		cpi		tmp0,1
		brmi	int_uart_put_PGM_str_loop
		lpm		tmp0,Z+					;load char
		tst		tmp0					;string terminated?
		breq	int_uart_put_exit
		rcall	int_uart_putchar
		rjmp	int_uart_put_PGM_str_loop
int_uart_put_exit:
		pop		tmp0
		ret


uart_int_fd_buff_cseg_end:				;cseg end address

.org	URXCaddr
		rjmp	ISR_URXC				;uart rx complete
.org	UDREaddr
		rjmp	ISR_UDRE				;uart data register empty
